@@ -80,9 +80,11 @@ struct CRCPBasic_C {
 //  Assignment should be done only by using basic_assign(). Before the variable
 //  goes out of scope, basic_free_stack() must be called.
 //
-//  For C, define a dummy struct with the right size, so that it can be
-//  allocated on the stack. For C++, the CRCPBasic is declared in cwrapper.cpp.
-#ifdef __cplusplus
+//  For downstream projects, define a dummy struct with the right size, so
+//  that it can be allocated on the stack. When building the library in
+//  cwrapper.cpp, the CRCPBasic is declared in cwrapper.cpp which removes
+//  the need to cast the C struct to C++ struct every time.
+#ifdef symengine_EXPORTS
 typedef struct CRCPBasic basic_struct;
 #else
 typedef struct CRCPBasic_C basic_struct;
@@ -352,6 +354,8 @@ CWRAPPER_OUTPUT_TYPE basic_log(basic s, const basic a);
 CWRAPPER_OUTPUT_TYPE basic_floor(basic s, const basic a);
 //! Assigns s = ceiling(a).
 CWRAPPER_OUTPUT_TYPE basic_ceiling(basic s, const basic a);
+//! Assigns s = sign(a).
+CWRAPPER_OUTPUT_TYPE basic_sign(basic s, const basic a);
 
 //! Assigns s = atan2(a, b).
 CWRAPPER_OUTPUT_TYPE basic_atan2(basic s, const basic a, const basic b);
@@ -775,6 +779,12 @@ CWRAPPER_OUTPUT_TYPE basic_evalf(basic s, const basic b, unsigned long bits,
 //! Wrapper for as_numer_denom
 CWRAPPER_OUTPUT_TYPE basic_as_numer_denom(basic numer, basic denom,
                                           const basic x);
+//! Wrapper for as_two_terms for expressions of type Add
+CWRAPPER_OUTPUT_TYPE basic_add_as_two_terms(basic term1, basic term2,
+                                            const basic s);
+//! Wrapper for as_two_terms for expressions of type Mul
+CWRAPPER_OUTPUT_TYPE basic_mul_as_two_terms(basic term1, basic term2,
+                                            const basic s);
 
 //! Wrapper for LambdaRealDoubleVisitor
 typedef struct CLambdaRealDoubleVisitor CLambdaRealDoubleVisitor;
