@@ -47,10 +47,34 @@ using std::make_unique;
 using llvm::make_unique;
 #endif
 
+#if (LLVM_VERSION_MAJOR >= 18)
+using CodeGenOptLevel = llvm::CodeGenOptLevel;
+#else
+using CodeGenOptLevel = llvm::CodeGenOpt::Level;
+#endif
+
+#if (LLVM_VERSION_MAJOR >= 20)
+auto GetDeclaration = [](llvm::Module *M, llvm::Intrinsic::ID id,
+                         llvm::ArrayRef<llvm::Type *> Tys) {
+    return llvm::Intrinsic::getOrInsertDeclaration(M, id, Tys);
+};
+#else
+const auto &GetDeclaration = llvm::Intrinsic::getDeclaration;
+#endif
+
+#if (LLVM_VERSION_MAJOR >= 21)
+#define AddNoCapture(A) A.addCapturesAttr(llvm::CaptureInfo::none())
+#else
+#define AddNoCapture(A) A.addAttribute(llvm::Attribute::NoCapture)
+#endif
+
 class IRBuilder : public llvm::IRBuilder<>
 {
 };
 
+LLVMVisitor::LLVMVisitor() = default;
+LLVMVisitor::~LLVMVisitor() = default;
+
 llvm::Value *LLVMVisitor::apply(const Basic &b)
 {
     b.accept(*this);
@@ -67,54 +91,33 @@ llvm::Function *LLVMVisitor::get_function_type(llvm::LLVMContext *context)
 {
     std::vector<llvm::Type *> inp;
     for (int i = 0; i < 2; i++) {
+#if (LLVM_VERSION_MAJOR >= 22)
+        inp.push_back(llvm::PointerType::get(*context, 0));
+#else
         inp.push_back(llvm::PointerType::get(get_float_type(context), 0));
+#endif
     }
     llvm::FunctionType *function_type = llvm::FunctionType::get(
         llvm::Type::getVoidTy(*context), inp, /*isVarArgs=*/false);
-    auto F = llvm::Function::Create(function_type,
-                                    llvm::Function::InternalLinkage, "", mod);
+    auto F = llvm::Function::Create(
+        function_type, llvm::Function::InternalLinkage, "symengine_func", mod);
     F->setCallingConv(llvm::CallingConv::C);
-#if (LLVM_VERSION_MAJOR < 5)
-    {
-        llvm::SmallVector<llvm::AttributeSet, 4> attrs;
-        llvm::AttributeSet PAS;
-        {
-            llvm::AttrBuilder B;
-            B.addAttribute(llvm::Attribute::ReadOnly);
-            B.addAttribute(llvm::Attribute::NoCapture);
-            PAS = llvm::AttributeSet::get(mod->getContext(), 1U, B);
-        }
-
-        attrs.push_back(PAS);
-        {
-            llvm::AttrBuilder B;
-            B.addAttribute(llvm::Attribute::NoCapture);
-            PAS = llvm::AttributeSet::get(mod->getContext(), 2U, B);
-        }
-
-        attrs.push_back(PAS);
-        {
-            llvm::AttrBuilder B;
-            B.addAttribute(llvm::Attribute::NoUnwind);
-            B.addAttribute(llvm::Attribute::UWTable);
-            PAS = llvm::AttributeSet::get(mod->getContext(), ~0U, B);
-        }
-
-        attrs.push_back(PAS);
-
-        F->setAttributes(llvm::AttributeSet::get(mod->getContext(), attrs));
-    }
-#else
     F->addParamAttr(0, llvm::Attribute::ReadOnly);
+#if (LLVM_VERSION_MAJOR >= 21)
+    F->addParamAttr(1, llvm::Attribute::getWithCaptureInfo(
+                           *context, llvm::CaptureInfo::none()));
+    F->addParamAttr(0, llvm::Attribute::getWithCaptureInfo(
+                           *context, llvm::CaptureInfo::none()));
+#else
     F->addParamAttr(0, llvm::Attribute::NoCapture);
     F->addParamAttr(1, llvm::Attribute::NoCapture);
+#endif
     F->addFnAttr(llvm::Attribute::NoUnwind);
 #if (LLVM_VERSION_MAJOR < 15)
     F->addFnAttr(llvm::Attribute::UWTable);
 #else
     F->addFnAttr(llvm::Attribute::getWithUWTableKind(
         *context, llvm::UWTableKind::Default));
-#endif
 #endif
     return F;
 }
@@ -126,7 +129,7 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     llvm::InitializeNativeTarget();
     llvm::InitializeNativeTargetAsmPrinter();
     llvm::InitializeNativeTargetAsmParser();
-    context = std::make_shared<llvm::LLVMContext>();
+    context = make_unique<llvm::LLVMContext>();
     symbols = inputs;
 
     // Create some module to put our function into it.
@@ -167,11 +170,7 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     }
 
     auto it = F->args().begin();
-#if (LLVM_VERSION_MAJOR < 5)
-    auto out = &(*(++it));
-#else
     auto out = &(*(it + 1));
-#endif
     std::vector<llvm::Value *> output_vals;
 
     if (symbolic_cse) {
@@ -210,13 +209,7 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     llvm::verifyFunction(*F, &llvm::outs());
 
     //     std::cout << "LLVM IR" << std::endl;
-    // #if (LLVM_VERSION_MAJOR < 5)
-    //     module->dump();
-    // #else
     //     module->print(llvm::errs(), nullptr);
-    // #endif
-
-    // module->print(llvm::errs(), nullptr);
 
     // Optimize the function using default passes from PassBuilder
     // FunctionSimplificationPipeline for the opt_level
@@ -244,26 +237,29 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     } else if (opt_level == 2) {
         pb_opt_level = OptimizationLevel::O2;
     }
+
+    if (opt_level != 0) {
 #if (LLVM_VERSION_MAJOR < 6)
-    FPM = PB.buildFunctionSimplificationPipeline(pb_opt_level);
+        FPM = PB.buildFunctionSimplificationPipeline(pb_opt_level);
 #elif (LLVM_VERSION_MAJOR < 12)
-    FPM = PB.buildFunctionSimplificationPipeline(
-        pb_opt_level, llvm::PassBuilder::ThinLTOPhase::None);
+        FPM = PB.buildFunctionSimplificationPipeline(
+            pb_opt_level, llvm::PassBuilder::ThinLTOPhase::None);
 #else
-    FPM = PB.buildFunctionSimplificationPipeline(
-        pb_opt_level, llvm::ThinOrFullLTOPhase::None);
+        FPM = PB.buildFunctionSimplificationPipeline(
+            pb_opt_level, llvm::ThinOrFullLTOPhase::None);
 #endif
-    FPM.run(*F, FAM);
+        FPM.run(*F, FAM);
+    }
 
     // std::cout << "Optimized LLVM IR" << std::endl;
     // module->print(llvm::errs(), nullptr);
 
     // Now we create the JIT.
     std::string error;
-    executionengine = std::shared_ptr<llvm::ExecutionEngine>(
+    executionengine = std::unique_ptr<llvm::ExecutionEngine>(
         llvm::EngineBuilder(std::move(module))
             .setEngineKind(llvm::EngineKind::Kind::JIT)
-            .setOptLevel(static_cast<llvm::CodeGenOpt::Level>(opt_level))
+            .setOptLevel(static_cast<CodeGenOptLevel>(opt_level))
             .setErrorStr(&error)
             .create());
 
@@ -272,7 +268,7 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     {
     public:
         std::string &ss_;
-        MemoryBufferRefCallback(std::string &ss) : ss_(ss) {}
+        explicit MemoryBufferRefCallback(std::string &ss) : ss_(ss) {}
 
         void notifyObjectCompiled(const llvm::Module *M,
                                   llvm::MemoryBufferRef obj) override
@@ -285,7 +281,7 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
         std::unique_ptr<llvm::MemoryBuffer>
         getObject(const llvm::Module *M) override
         {
-            return NULL;
+            return nullptr;
         }
     };
 
@@ -301,6 +297,9 @@ void LLVMVisitor::init(const vec_basic &inputs, const vec_basic &outputs,
     symbols.clear();
 }
 
+LLVMDoubleVisitor::LLVMDoubleVisitor() = default;
+LLVMDoubleVisitor::~LLVMDoubleVisitor() = default;
+
 double LLVMDoubleVisitor::call(const std::vector<double> &vec) const
 {
     double ret;
@@ -330,6 +329,9 @@ void LLVMLongDoubleVisitor::call(long double *outs,
 }
 #endif
 
+LLVMFloatVisitor::LLVMFloatVisitor() = default;
+LLVMFloatVisitor::~LLVMFloatVisitor() = default;
+
 float LLVMFloatVisitor::call(const std::vector<float> &vec) const
 {
     float ret;
@@ -354,6 +356,10 @@ void LLVMVisitor::bvisit(const Integer &x)
 }
 
 #ifdef SYMENGINE_HAVE_LLVM_LONG_DOUBLE
+
+LLVMLongDoubleVisitor::LLVMLongDoubleVisitor() = default;
+LLVMLongDoubleVisitor::~LLVMLongDoubleVisitor() = default;
+
 void LLVMLongDoubleVisitor::convert_from_mpfr(const Basic &x)
 {
 #ifndef HAVE_SYMENGINE_MPFR
@@ -461,15 +467,14 @@ llvm::Function *LLVMVisitor::get_powi()
 #if (LLVM_VERSION_MAJOR > 12)
     arg_type.push_back(llvm::Type::getInt32Ty(mod->getContext()));
 #endif
-    return llvm::Intrinsic::getDeclaration(mod, llvm::Intrinsic::powi,
-                                           arg_type);
+    return GetDeclaration(mod, llvm::Intrinsic::powi, arg_type);
 }
 
 llvm::Function *get_float_intrinsic(llvm::Type *type, llvm::Intrinsic::ID id,
                                     unsigned n, llvm::Module *mod)
 {
     std::vector<llvm::Type *> arg_type(n, type);
-    return llvm::Intrinsic::getDeclaration(mod, id, arg_type);
+    return GetDeclaration(mod, id, arg_type);
 }
 
 void LLVMVisitor::bvisit(const Pow &x)
@@ -887,25 +892,7 @@ llvm::Function *LLVMVisitor::get_external_function(const std::string &name,
             func_type, llvm::GlobalValue::ExternalLinkage, name, mod);
         func->setCallingConv(llvm::CallingConv::C);
     }
-#if (LLVM_VERSION_MAJOR < 5)
-    llvm::AttributeSet func_attr_set;
-    {
-        llvm::SmallVector<llvm::AttributeSet, 4> attrs;
-        llvm::AttributeSet attr_set;
-        {
-            llvm::AttrBuilder attr_builder;
-            attr_builder.addAttribute(llvm::Attribute::NoUnwind);
-            attr_set
-                = llvm::AttributeSet::get(mod->getContext(), ~0U, attr_builder);
-        }
-
-        attrs.push_back(attr_set);
-        func_attr_set = llvm::AttributeSet::get(mod->getContext(), attrs);
-    }
-    func->setAttributes(func_attr_set);
-#else
     func->addFnAttr(llvm::Attribute::NoUnwind);
-#endif
     return func;
 }
 
@@ -937,7 +924,7 @@ void LLVMVisitor::loads(const std::string &s)
     llvm::InitializeNativeTarget();
     llvm::InitializeNativeTargetAsmPrinter();
     llvm::InitializeNativeTargetAsmParser();
-    context = std::make_shared<llvm::LLVMContext>();
+    context = make_unique<llvm::LLVMContext>();
 
     // Create some module to put our function into it.
     std::unique_ptr<llvm::Module> module
@@ -952,10 +939,10 @@ void LLVMVisitor::loads(const std::string &s)
     auto F = get_function_type(context.get());
 
     std::string error;
-    executionengine = std::shared_ptr<llvm::ExecutionEngine>(
+    executionengine = std::unique_ptr<llvm::ExecutionEngine>(
         llvm::EngineBuilder(std::move(module))
             .setEngineKind(llvm::EngineKind::Kind::JIT)
-            .setOptLevel(llvm::CodeGenOpt::Level::Aggressive)
+            .setOptLevel(CodeGenOptLevel::Aggressive)
             .setErrorStr(&error)
             .create());
 
