@@ -22,15 +22,17 @@ void print_rational_class(const rational_class &r, std::ostringstream &s)
 
 void LatexPrinter::bvisit(const Symbol &x)
 {
-    std::string name = x.get_name();
+    str_ = _print_symbol(x.get_name());
+}
 
+std::string LatexPrinter::_print_symbol(const std::string &name)
+{
     if (name.find('\\') != std::string::npos
         or name.find('{') != std::string::npos) {
-        str_ = name;
-        return;
+        return name;
     }
     if (name[0] == '_') {
-        name = name.substr(1, name.size());
+        return _print_symbol(name.substr(1, name.size() - 1));
     }
     std::vector<std::string> greeks
         = {"alpha",  "beta",  "gamma", "Gamma", "delta",   "Delta",   "epsilon",
@@ -41,16 +43,21 @@ void LatexPrinter::bvisit(const Symbol &x)
 
     for (auto &letter : greeks) {
         if (name == letter) {
-            str_ = "\\" + name;
-            return;
-        }
-        if (name.size() > letter.size() and name.find(letter + "_") == 0) {
-            str_ = "\\" + name;
-            return;
+            return "\\" + name;
         }
     }
-    str_ = name;
-    return;
+
+    size_t idx = name.find("_");
+    if (idx == std::string::npos || idx == name.size() - 1) {
+        return name;
+    } else if (idx < name.size() - 2) {
+        return _print_symbol(name.substr(0, idx)) + "_{"
+               + _print_symbol(name.substr(idx + 1, name.size() - idx - 1))
+               + "}";
+    } else {
+        return _print_symbol(name.substr(0, idx)) + "_"
+               + name.substr(idx + 1, name.size() - idx - 1);
+    }
 }
 
 void LatexPrinter::bvisit(const Rational &x)
@@ -547,7 +554,7 @@ void LatexPrinter::bvisit(const Ceiling &x)
 void LatexPrinter::bvisit(const Abs &x)
 {
     std::ostringstream o;
-    o << "\\left|" << apply(x.get_arg()) << "}\\right|";
+    o << "\\left|" << apply(x.get_arg()) << "\\right|";
     str_ = o.str();
 }
 
