@@ -1,3 +1,6 @@
+#ifndef SYMENGINE_STRINGBOX_H
+#define SYMENGINE_STRINGBOX_H
+
 #include <string>
 #include <vector>
 
@@ -51,3 +54,5 @@ public:
 };
 
 }; // namespace SymEngine
+
+#endif
