@@ -9,7 +9,7 @@
 namespace SymEngine
 {
 //! Less operator `(<)` using cmp:
-struct PrinterBasicCmp {
+struct PrinterBasicCmpUnicode {
     //! true if `x < y`, false otherwise
     bool operator()(const RCP<const Basic> &x, const RCP<const Basic> &y) const
     {
@@ -327,6 +327,19 @@ void UnicodePrinter::bvisit(const Union &x)
     box_ = box;
 }
 
+void UnicodePrinter::bvisit(const Intersection &x)
+{
+    auto container = x.get_container();
+    StringBox box = apply(*container.begin());
+    StringBox op(U8(" \u2229 "), 3);
+    for (auto it = ++(container.begin()); it != container.end(); ++it) {
+        box.add_right(op);
+        StringBox next = apply(*it);
+        box.add_right(next);
+    }
+    box_ = box;
+}
+
 void UnicodePrinter::bvisit(const Complement &x)
 {
     StringBox box = apply(*x.get_universe());
@@ -385,7 +398,7 @@ void UnicodePrinter::bvisit(const Add &x)
 {
     StringBox box;
     bool first = true;
-    std::map<RCP<const Basic>, RCP<const Number>, PrinterBasicCmp> dict(
+    std::map<RCP<const Basic>, RCP<const Number>, PrinterBasicCmpUnicode> dict(
         x.get_dict().begin(), x.get_dict().end());
 
     if (neq(*(x.get_coef()), *zero)) {
