@@ -130,7 +130,9 @@ void CSRMatrix::set(unsigned i, unsigned j, const RCP<const Basic> &e)
                 k++;
             }
             break;
-        } else if (j_[mid] >= j and j_[mid - 1] < j) {
+        }
+        SYMENGINE_ASSERT(mid > 0);
+        if (j_[mid] >= j and j_[mid - 1] < j) {
             k = mid;
             break;
         } else if (j_[mid - 1] >= j) {
@@ -159,6 +161,19 @@ void CSRMatrix::set(unsigned i, unsigned j, const RCP<const Basic> &e)
     }
 }
 
+tribool CSRMatrix::is_real(const Assumptions *assumptions) const
+{
+    RealVisitor visitor(assumptions);
+    tribool cur = tribool::tritrue;
+    for (auto &e : x_) {
+        cur = and_tribool(cur, visitor.apply(*e));
+        if (is_false(cur)) {
+            return cur;
+        }
+    }
+    return cur;
+}
+
 unsigned CSRMatrix::rank() const
 {
     throw NotImplementedError("Not Implemented");
@@ -404,7 +419,7 @@ bool CSRMatrix::csr_has_sorted_indices(const std::vector<unsigned> &p_,
                                        unsigned row_)
 {
     for (unsigned i = 0; i < row_; i++) {
-        for (unsigned jj = p_[i]; jj < p_[i + 1] - 1; jj++) {
+        for (unsigned jj = p_[i]; jj + 1 < p_[i + 1]; jj++) {
             if (j_[jj] > j_[jj + 1])
                 return false;
         }
@@ -630,6 +645,7 @@ void csr_diagonal(const CSRMatrix &A, DenseMatrix &D)
             } else if (A.j_[jj] < i) {
                 row_start = jj + 1;
             } else {
+                SYMENGINE_ASSERT(jj > 0);
                 row_end = jj - 1;
             }
         }
